#ifndef MYLIB_H
#define MYLIB_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Prints a hello message from the framework
 */
void mylib_hello(void);

#ifdef __cplusplus
}
#endif

#endif // MYLIB_H
